const char *version_string="2014-06-26";
// #define CFG_MPI        /* define CFG_MPI + mpicc ... */ 
/*
 measure speed of memory (best and worst case)
 also FDIR
  - Ausfaelle erkennen (fault-detection)          
  - Ausfaelle einkreisen und bestimmen (fault-isolation)
  - geeignete Korrekturen (recovery)    
 
 INST_SEQ: # single task (no parallelism)
          gcc -Wall -O2 -o memspeed memspeed.c
          # memory footprint: 
 INST_MPI: # OpenMPI-usage (parallel on distributed memory)
          mpicc -Wall -O2 -o memspeed memspeed.c
          # optional: mpicc|gcc -DCFG_OMP -fopenmp ...
          #   cpuset is a hex bit pattern for used cpu cores (p.e. 0x05,0x0f)
          #     useful to bind tasks to cores using sched_setaffinity()
          #   use cc or gcc if mpicc is not availabe
          mpirun -np <tasks> ./memspeed --log2size=<log2size> --mintime=<sec>
          # needs: memory = tasks*2^log2size Bytes
          #        time   = (1..2)*2*mintime*(log2size-8) seconds
          mpirun -np 8 --bynode ./memspeed #   
          # runtime is about 2*(minsize...2minsize)*(log2size-7)
 INST_OMP:# OpenMP-usage (parallel on shared memory)
          yum install libgomp
          gcc -Wall -O2 -DCFG_OMP -fopenmp -o memspeed memspeed.c
          export OMP_WAIT_POLICY="PASSIVE" # OpenMP-3.0 waits saving power 
          export OMP_NUM_THREADS=8         # OpenMP     8 threads
          OMP_PROC_BIND=TRUE # thread bounding, thread affinity
          OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./memspeed
          # memory footprint: 
 MONITOR: ps -Lf     # UID PID PPID LWP(!) C NLWP STIME TTY TIME CMD
          ps -T      # PID SPID TIME TTY CMD
          top -H
          cat /sys/devices/system/edac/mc/mc[0-3]/ce_count | xargs # ECC-Errors
          cat /sys/devices/system/edac/mc/poll_msec # 1000 ms # x86+amd
 TEST-Suites:
          kautz2009-2012 lot of ECC-Errors and types (CE,UE,panics)
          comp2-ECC-CEs+Crashs

ToDo: better read random index from array?
 - reduce code size and memory footprint (90M-180M openmpi), mpi-less?
 - check FLOPs for mem+power stress using oprofile
 - modules speed random, random_prefetch, seq, r[w], + stress + mark badmem
   MEMstress +FPU+GPU+DSK20W+DVD+USB2W+ETH+IB(=PSU)-stress? vs. Verbrauch?
   test: stresstypes vs. power on comp2
   see badRAM-linux-patch + memtest86-tests(random+mod20)
    output speed only if changed, not after 10 changes (less output) +timestamps
 - add OpenMP threads (alternative: fork/clone+pipe or pthreads)
ChangeLog:
 2014-06-26: more help, NA controls prefetching of random access now

 # P4-2.6GHz/2_maxspeed=(1.3GHz/(loop=5Instr)=260MHz)*(long=4)=1040MB/s
 # see also memspeed.gpl
 # compare with numademo 8000000000 # 8GB per run? (drop_caches! 5 vs. 8GB/s)
 #    2012comp2 memset=5-8GB/s 
 # System               Array Rand MaxMB/s Err  # new 4x->4summed
 Nehalem-2GHz-4MB       256MB   70   6338       114ns 64bit-gcc-4.3.3 72GB -O2
 Nehalem-2GHz-4MB       256MB   70   6340       114ns 64bit-gcc-4.3.3 72GB -O1
 Nehalem-2GHz-4MB       256MB   68   1850       118ns 64bit-gcc-4.3.3 72GB -O0 more x86-instr.
 Nehalem-2GHz-4MB    2x 256MB   66   4707       122ns 64bit-gcc-4.3.3 72GB -O0 more x86-instr.
 PM- 600MHz-1MB         512kB  284    393    -   14ns 32bit-gcc-4.1.2 2GB hdparm=385MB/s 0808
 PM- 600MHz-1MB         256MB   23    382    -  174ns 32bit-gcc-4.1.2 2GB hdparm=385MB/s 0808
 PM-1400MHz-1MB          64MB   31    870    -  127ns 32bit-gcc-4.1.2 2GB
 P4-2.6GHz/2-512kB      256MB   20   1090    -  197ns 32bit-gcc-4.1.1 hdparm=1181MB/s FSB=4*100MHz Mem=266MHz*8=2128MB/s CPU=13*100MHz (default=13*200MHz) L2=512k=2^19 0808
 P4-2.6GHz/2-512kB HT 2x256MB   26   1533    -  307ns 32bit-gcc-4.1.1 hdparm=1181MB/s FSB=4*100MHz Mem=266MHz*8=2128MB/s CPU=13*100MHz (default=13*200MHz) L2=512k=2^19 0808
 SunFire-880             32MB   23    220    6  346ns 64bit-gcc-4.1.1 USparcIII 750MHz
 SunFire-880             32MB   23    232    7  333ns 64bit-CC-5.3-fast-xarch=v9
 SunFire-V490            32MB   92    798   19   84ns 64bit-gcc-4.1.1 USparcIV+ 10*150MHz L2=32MB 8-way
 DOpteron275-2.2GHz      32MB   89   2130   34   88ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron275-2.2GHz  2x  32MB  177   4129   16   88ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore 4*44MB/s+2*85MB/s
 DOpteron275-2.2GHz  4x  32MB  340   7156   92   91ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron275-2.2GHz  6x  32MB  346   7200   98  177ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore 4*44MB/s+2*85MB/s
 DOpteron275-2.2GHz  8x  32MB  332   7248   92  189ns 64bit-gcc-4.1.1 L2=1MB  hdparm-T=1945MB/s 2.6.18 2DualCore
 DOpteron885-2.6GHz      32MB   35    838   21  223ns 64bit-gcc-4.1.0 8user   hdparm-T=1660MB/s BW=6.4GB/s(1CPU)
 4HOpteron8431-2.4GHz 8x 32MB 8*79 8*1820    -  101ns 64bit-gcc-4.1.2 3user   hdparm-T=5800MB/s (summed)
 Xeon-2333Mz-4MB         32MB   69   3295       116ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     4x  32MB 4*62 4*1466       131ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     8x  32MB 8*55  8*734       144ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad 
 Xeon-2333Mz-4MB     1x 512MB   60   3269       133ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad
 Xeon-2333Mz-4MB     8x 512MB 8*39  8*735       204ns 64bit-gcc-4.1.1-long thor model=15 stepping=4  i5000P (maxBW=21GB/s) 2Quad  Aug08
 alpha-gs160-731MHz      16MB   20    295   49  397ns 64bit-gcc-4.1.1-long
 alpha-gs160-731MHz      16MB   20    295   49  397ns 64bit-cxx-6.5-long
 alpha-gs1280            32MB   31    453   50      - 64bit-cxx-6.5-long   (32cpus_per_rad) 0user
 alpha-gs1280            32MB   46    644   22  170ns 64bit-cxx-6.5-long   (8cpus_per_rad) 8user
 alpha-gs1280            32MB   46    339  170  170ns 64bit-gcc-4.1.1-long (8cpus_per_rad) 8user
 alpha-gs1280            32MB   77   1109    9  101ns 64bit-cxx-6.5-long   (1cpus_per_rad) 0user
 alpha-gs1280         32x32MB 2432  14400 1600      - 64bit-cxx-6.5-long   (32cpus_per_rad) 0user
 alpha-gs1280         32x32MB 2432  35520  640  102ns 64bit-cxx-6.5-long   (1cpus_per_rad) 0user
 alpha-es45-1250MHz      32MB   38    462   56  204ns 64bit-cxx-6.5-long 4user
 altix330-IA64-1.5GHz    32MB   22    309   12  349ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz 8x 32MB  168   2304   80  970ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz 15x32MB  660   9300  300  176ns 64bit-gcc-4.1.1 Apr07
 altix330-IA64-1.5GHz    32MB   54    670   14  143ns numactl --localalloc
 altix330-IA64-1.5GHz 8x 32MB  392   4760  360  158ns numactl --localalloc
 altix330-IA64-1.5GHz 15x32MB  735  10050  300  158ns numactl --localalloc
 altix330-IA64-1.5GHz 8x512MB  418   5750       153ns min= 8*52MB/s aff+local
 altix330-IA64-1.5GHz 8x128MB  427   5748       150ns min= 8*53MB/s aff+local
 altix330-IA64-1.5GHz 2x128MB   91    709       174ns min= 2*45MB/s c0,1+local
 altix330-IA64-1.5GHz 2x128MB  107   1438       150ns min= 2*53MB/s aff0,8
 altix330-IA64-1.5GHz 2x512MB   89    710       179ns min= 2*45MB/s c0,1+local
 altix330-IA64-1.5G  16x128MB  448   6928       285ns min=16*28MB/s localalloc
 altix330-IA64-1.5G  16x128MB  700  10800       182ns min=16*44MB/s localalloc
 altix330-IA64-1.5G  16x128MB  729  11335       182ns min=16*45MB/s aff+lalloc
 altix330-IA64-1.5G  16x512MB  480   7456       264ns min=16*30MB/s localalloc
# echo 1 > /proc/sys/vm/drop_caches + numademo 1024m # ???
# 30*8quantum numademo 1024m 1node  stream=3.7GB/s     rnd=137MB/s
# 22*16comp2  numademo 1024m 4nodes stream=2.8-6.3GB/s rnd=93-125MB/s 4*16GB
# check with linkstat!
# ----------- old data (step only) --------------------
 System                 Array Min MaxMB/s Err ln2MinStep # ---- old(!) ----
 PM- 600MHz-1MB 512MB    16MB   52    364   16    12   
 PM-1400MHz-1MB 512MB    16MB   50    364    8    12   
 Xeon-3GHz-1MB  2GB     256MB   39   1800   48    16   32bit-gcc-2.96  connect model=3 stepping=4  hdparm=800MB/s  busclock=200MHz
 Xeon-3GHz-2MB  2GB     256MB   44   2700   57    20   32bit-gcc-4.0.1 bb    model=4 stepping=3  hdparm=1472MB/s
 Xeon-3GHz-2MB 12GB     256MB   39   1422   31    20   64bit-gcc-4.1.1 morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 Xeon-3GHz-2MB 12GB     512MB   38   1450   31    20   32bit-gcc-4.1.1-long morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 Xeon-3GHz-2MB 12GB    8192MB   80   1170   23    19   64bit-gcc-4.1.1-long morus model=4 stepping=10 memtest86+1.65=1257MB/s hdparm=1286MB/s
 alpha-gs1280 128GB      32MB   36   1117   39    15   64bit-gcc-3.4.3-long (32cpus_per_rad) 2/82ns==25MB/s
 alpha-gs1280 128GB      32MB   41   2485   45    15   64bit-cxx-6.5-long   (32cpus_per_rad) (32x2.5GB/s=80GB/s local) remote.maxBW=6.3GB/s local.maxBW=12.3GB/s(767MHz*8channels*2Byte) 8B/(82ns..250ns)=>32..98MB/s
 alpha-gs1280 128GB      32MB   67   1117   39    16   64bit-gcc-3.4.3-long (1cpus_per_rad) 2/82ns==25MB/s
 alpha-gs1280 128GB      32MB   61   2485   47    16   64bit-cxx-6.5-long   (1cpus_per_rad) (32x2.5GB/s=80GB/s local)
 alpha-gs1280 128GB      16GB   24    572    7    23   64bit-cxx-6.5-long   (8cpus_per_rad)
 alpha-gs1280 128GB      16GB   29   1094   24    23   64bit-gcc-4.1.1-long (8cpus_per_rad)
 alpha-es45-1250MHz      16MB   87    818   20    11   cxx-int? 4/5CPU
 DOpteron275-2.2GHz-1MB  16MB   35   1500   34    13   64bit-gcc-3.4.6      hdparm-T=1900MB/s
 DOpteron885-2.6GHz-1MB 256MB   38   1120   39    15   64bit-gcc-4.0.3      hdparm-T=1660MB/s BW=6.4GB/s(1CPU)
 DOpteron885-2.6GHz-1MB 512MB   68   2300   82    15   64bit-gcc-4.0.3-long hdparm-T=1660MB/s
 DOpteron885-2.6GHz-1MB 512MB*4 46   2220   77    15   CPU3-6! 64bit-gcc-4.0.3-long hdparm-T=1660MB/s
 MinMax fuer 2 Jobs etc. # linkstat?
 cpu.model=generation stepping=version (same masque)
 ToDo: try __*_prefetch and improve loop (tune for generated asm-code)
 Theory vs. Samples (yum install edac-utils):
   NumCPUs*Memchannels/CPU*maxDatarate vs. DIMMs/Channel*Ranks/DIMM*Speed
 see ct23/2013 REP MOVSB different speeds, bugs (break loop, INT,VMs,...)
 Memory-Footprint: openmpi is bad! 128MB/node (tune ompi-compilation!)
 comp2: 2*AMD-6134*4Channels * max 3 DIMMs/channel = 2*42.7GB/s bei 1DIMM/ch
        2*8GB/Channel (DDR3-1333) = 10.664GB/s / channel = 85.312GB/s
          aBW=22.3GB/s? (unroll2?=26.0GB/s)  memtest4.20=3746MB/s(singleCPU)
          aBW=46.6GB/sR 2013-06
           http://en.wikipedia.org/wiki/Memory_footprint
           1MBmemusage: approx 128MB sm for openmpi
            --mca btl self,sm,tcp # 1*6M/126M ... 4*6M/191MB... 8*7M/193MB
              67MB tmp/openmpi-sessions-jschulen@comp2_0/60576/1/shared_mem_pool.comp2
            --mca btl self,tcp # 8*5M/91MB (9MB openmpi-libs)
            single-task-mpi: 2M/23MB 9MB-ompi-lib
            single-task-nompi: 1.5M/5MB 2MB-libc /proc/$PID/smaps
     worst random at mem/core/2 = 4GB 
        test 1GB: 1*=40ns/8B 2*  16*155MB/s=52ns/8B/16 ca. 40 cycles
        test 4GB:                16*136MB/s=59ns/8B/16
 quantum: 2*E5450 (8GB/s?)
          8 DDR2-667(PC-5300F)=5.3GB/s(10.6GB/s DualChannel)
          aBW=6.17GB/sRW   
 pse:   2*6core-iX5650-2.66GHz = 2*3 Speicherkanaele(DDR3-1333) max 2*32GB/s
        18*16GB-DDR3-1066MHz(a8.5GB/s) = 3 DIMMs/Channel max 51GB/s (ToDo)
          aBW=29.5GB/sRW 
 quanta1: 4*8 L1=32kB(0.5-4ns) L2=4*8*256kB(0.7-10ns) L3=16MB(1.05-33ns) mem(1.44-128ns)
          aBW1=5.5GB/s rand1=63MB/s(127ns,760MB)
  # Tasks Mem/Task[B] loops t[s]  aBW[B/s] BW[B/s] lat[ns]
  1 7.73e+11         7 3110.84 3.48e+09 3.48e+09   4.60 lr+w4*+16
  1 7.73e+11         7 3080.90 3.51e+09 3.51e+09   4.55 lr+w4*+16
  1 7.73e+11         7 3042.04 3.56e+09 3.56e+09   4.50 lr+w4*+16
ToDo:
 - include memtest.c
 - add pthread
 - memory-footprints (pure,mpi-???,ompi-???,stst,dyn)
  memusage(ELF64+libc6+1MB+test32) method: (ulimit -v ....;./a.out) 2013-09
  gcc4.1.2                   4.8MB       
  gcc4.1.2+static            1.9MB       
  gcc4.1.2+openmp2.5        11.2MB(1) 21.5MB(2) 31.7MB(3) 42.0MB(4) 82.9MB(8)
  gcc4.1.2+openmp2.5-static  2.0MB(1) 12.2MB(2) 22.4MB(3) 32.7MB(4) 73.7MB(8)
  openmpi2.1+gcc4.1.2        90MB(1)   n*155MB(2-8)      
  gcc4.1.2+-lpthread         6.9MB(1) 
  # RESUME2013: -fopenmp +10MB*n overhead! ompi=+154MB*n overhead
 - messe auch free-difference or max malloc (shared LIBS!)
 - einzelprogramme besser als OMPI!?
 - audiotxt verbessern arecord als stream! fuer const. mem on PC
         
*/


#define _GNU_SOURCE 1 // this defines __USE_GNU in features.h for sched.h

#ifdef CFG_OMP
#include <omp.h>    /* yum install libgomp; gcc -fopenmp ... */
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* 2013-05 adding strncmp() */
#include <time.h>
#include <unistd.h> /* usleep */

#ifndef CFG_MPI    // for the case of MPI is not available
/* last tests: 2013-05-02 */
#include <sys/time.h>
double wtime(){ struct timeval tv;
 if(gettimeofday(&tv, NULL)){perror("timeofday");exit(2);}
 return tv.tv_sec+1e-6*tv.tv_usec; }
#define MPI_Init(a,b) 1
#define MPI_SUCCESS 1
#define MPI_Comm_size(MPI_COMM_WORLD, mpi_n)
#define MPI_Comm_rank(MPI_COMM_WORLD, mpi_i)
#define MPI_Barrier(MPI_COMM_WORLD)
#define MPI_Wtime()  wtime() /* 2013-05-02 itime replaced by wtime */
#define MPI_Finalize()
#define MPI_Abort(MPI_COMM_WORLD,xx) exit(xx)
#define MPI_Bcast(pt, len, MPI_DOUBLE, id, MPI_COMM_WORLD)
/* Reduce replacement works for MIN MAX SUM,
   Reduce to rank0, Allreduce to all ranks */
#define MPI_Reduce(snd, rec, SIZE, M_INT, M_MIN, rk, MCW) *rec=*snd
#define MPI_Allreduce(snd, rec, SIZE, M_INT, M_MIN, MCW) *rec=*snd
#else
#include <mpi.h>
#endif
#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
# include <unistd.h>  // declare sleep()+usleep
# include <sched.h>   // declare sched_setaffinity()
# ifndef _FEATURES_H  
#  warning "_FEATURES_H is not defined, try top add #include <features.h>"
# endif
# ifndef __USE_GNU                                                 
#  warning "expect __USE_GNU defined, but it isn't, try -D_GNU_SOURCE=1"
# endif
#endif


int main(int argn, char *argv[]){
  double t1, t2, tmin=1.0;
  int i1, 
      omp_n=1,          /* OpenMP omp_num_threads() */
      mpi_n=1, mpi_i=0, /* MPI threads and thread number */
      testmode=  5, /* 1=seq_read 2=pf_random 4=nonpf_random 8=stress ... */
      log2long;     // 5 for 2^5=int32 or 6 for 2^6=int64 (long int *aa)
  long int loops, i, j, a2, k,
     xy=0; // dummy vars for the innerst loop (unique names!)
  long int memhole0=0, memhole1=0; // memory hole (bad memory) 2013-11-07
  if (MPI_Init(&argn, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  setvbuf(stdout, NULL, _IONBF, 0); /* dont buffer output */
  if (!mpi_i)
   printf("# memspeed %s (c) 2008-2014 GPLv2 Joerg Schulenburg ovgu.de\n",
           version_string);
  /* 2013-05 add option strings --... for more flexibility */
  for (i1=1;i1<argn;i1++) { /* check all args */
    if (strncmp(argv[i1],"--mintime=",10)==0) { 
      tmin=atof(argv[i1]+10); if (tmin<0.0) tmin=0.0; continue; }
    if (strncmp(argv[i1],"--test=",7)==0) { 
      testmode =atoi(argv[i1]+7); continue; }
    if (!mpi_i) {
     printf("ERROR: bad or unknown options or arguments\n"
           " use: ./cpuspeed --mintime=1 --test=5\n");
     fputs("# --test=32   = power stress, max FLOPs (noMem)\n",stdout);
    }
    usleep(100000l); /* give 100ms for output */ MPI_Abort(MPI_COMM_WORLD,1);
  }
  /* get the log2size of long int, log2long-3 is log2(Bytesize(long)) */
  for (log2long=0; 1<<log2long < 8*sizeof(long int); log2long++);

  if (!mpi_i) printf("#");
   /* 2013-05 gcc -fopenmp -v ... =
     -L/usr/lib/gcc/x86_64-redhat-linux/4.1.2 -lgomp -lpthread */
#ifdef MPI_VERSION
  if (!mpi_i) printf(" MPI-%d.%d",MPI_VERSION,MPI_SUBVERSION);
#endif
  if (!mpi_i)
    printf(" mpi=%d int%d  mintime/s=%.1f test=%d\n",
      mpi_n, 1<<log2long, tmin, testmode);

  if ((memhole0 || memhole1) && (memhole1<=memhole0 || memhole0<0)){
    printf("ERROR: memhole end<=start or start<0\n");exit(1);}

#ifndef NA      /* use option -DNA=5 to modify inner loops */
#define NA (8)  /* vector array */
#endif


/* FLOP_t may be defined outside as float,double,int,short,char ... */
#ifdef FLOP_t
#  if FLOP_t != float && FLOP_t != double && FLOP_t != short
#    error "FLOP_t must be one of float, double or short" 
#  endif
#endif
#ifndef FLOP_t  /* double precission (DP) OR single precision (SP, 32bit) */
#define FLOP_t long int
#endif
/* ToDo: FLOAT64-STRESS = max power consumption 2013-09-12 */
//if (testmode&32) 
  {  // float precision may double peak performance
     FLOP_t ra[NA+16],// __attribute__ ((aligned (16))), // not needed  NP
            rb[NA+16],// __attribute__ ((aligned (16))); // NV
            rc[NA+16],rd[40];// __attribute__ ((aligned (4))); // NV
  int tmode=32, cnt=0;
  if (!mpi_i) {
    printf("# CPU"
#ifdef __GNUC__
     "-gcc-%d.%d"
#endif
     "-features:"
#ifdef __MMX__   // 64bit
     " MMX"  // pentium-mmx pentium2 c3 (MM0..MM7 64bit=1i64,2i32,4i16,8i8)
      // PADDB, PADDD, PADDSB, PADDSW, PADDUSB, PADDUSW, PADDW, 
      // PMADDWD, PMULHW, PMULLW
#endif
#ifdef __SSE__  // 8reg128bit, 2 operands (float only)
     " SSE"  // pentium3  XMM0..XMM7 128bit=4f32  ADDPS, ADDSS, MULPS, MULSS
     // float a[4],b[4]  a[]+=b[] 4ADDs/clk
#endif
#ifdef __SSE2__ // +integer (ToDo: check asm output by -S -msse2 -mno-sse2)
     " SSE2"  // pentium-m pentium4  XMM+2f64,16i8,8i16,4i32,2i64
     // ADDPD,MULPD 2f64 ADDSD,MUSD 1f64  (see .S output)  PADDQ PMULUDQ
#endif
#ifdef __SSE3__
     " SSE3"  // complex-arith ADDSUBPD ...PS 
#endif
#ifdef __SSSE3__
     " SSSE3"  // atom core2  PMULHRSW, PMADDUBSW
#endif
#ifdef __SSE4a__
     " SSE4a"  // core2++ PMULLD PMULDQ
#endif
#ifdef __3dNOW__
     " 3dNOW"   // k8  mmx+2f32  PFADD, PFMUL
#endif
#ifdef __FMA__
     " FMA"     // FMA fused multiply add xmm0..3
#endif
#ifdef __AVX__  
     " AVX"     // 256bit, 3 operands, needs OS-support a[]=b[]+c[] GCC4.6
    // only low128bit for integer usable!
#endif
#ifdef __AVX2__  // ???
     " AVX2"     // 256bit-SSE-iOPs, 256bit for integers, +FMA3(3register)
#endif
    "\n"
#ifdef __GNUC__
     , __GNUC__ ,  __GNUC_MINOR__
#endif
    );
    printf("# FLOAT%d-Power-Stress (mode=%d, 2DPFLOP/clk @ X5650)\n"
   "# mpi omp loops t[s]  aFLOPS[1/s] FLOPS[1/s] lat[ns]\n", 
   (int)sizeof(FLOP_t)*8, testmode);
  }
  {
    int k2;
    for (k2=0;k2<NA;k2++) ra[k2]=+k2+777; // p.e. +0.25
    for (k2=0;k2<NA;k2++) rb[k2]=-k2-777; // p.e. -0.25
    for (k2=0;k2<40;k2++) rd[k2]=k2;
    rb[0]=0x55555555l; // 32bit mask0
    rb[1]=0x11111111l; // 32bit mask1
    rb[2]=0x01010101l; // 32bit mask2
    rb[3]=0x00010001l; // 32bit mask3
    rb[4]=0x00000001l; // 32bit mask4 = rot32
    rb[5]=0x00000000l; // 32bit mask5
    rb[6]=0x00000000l; // 32bit mask6
    // not fix at compile time! cant be optimized away
    if (argn>8+NA) {
      for (k2=0;k2<NA  ;k2++) ra[k2]=(long int)(atoi(argv[k2+1])); // p.e. +0.25
      for (k2=0;k2<NA+8;k2++) rb[k2]=(long int)(atoi(argv[k2+1])); // p.e. -0.25
    }
  }
  // measure memory throughput for step0
  for (cnt=0;;cnt++) { // do not reduce msgsize, stress it endless! alen=1<<alog
    /* rot through bit 8 + 16 until 1 bit found */
    a2=64000000l; /* fix for unroll */
    loops=0;
    t1=MPI_Wtime();
    for (i=1;i<(1l<<30);i*=2) {
      int k2;
      // for (k2=0;k2<NA;k2++) ra[k2]=k2; // nonzero! 
      j=i; // 2013 comp2: first round runs slower, autotune cache method?
      // 1st_round=393MB/s, 2nd++_round=1610MB/s (faktor4), training? throttle?
      do { // repeat until some seconds are reached
         // k=a2/2;do{...}while(--k);  # 8.49ns(256MB) 0.94GFLOPs(19.6%)
// 3DNow (AMD deprec.) = packed-2xfloat32-SIMD/VECT  2..4FLOPs +AVG8i8
// 3dnowprefetch = prefetch 32B to L1, multithread.monitor+mwait
// sse2=SIMD-int
// sse3=complex, add(mltply) 2f64 or 2*2f32 in same reg. (=3DNow) 2..4FLOP
// sse4a=LZCNT, POPCNT,   
// Intel Core 2 and Nehalem, SSE:      4 DP FLOPs/cycle ( 8 SP FLOPs/cycle)
// Intel Sandy Bridge/Ivy Bridge, AVX: 8 DP FLOPs/cycle (16 SP FLOPs/cycle)
// Intel Haswell with AVX2/FMA:       16 DP FLOPs/cycle (32 SP FLOPs/cycle)
// AMD K10 (Opteron, Athlon etc) SSE2: 4 DP FLOPs/cycle ( 8 SP FLOPs/cycle)
// AMD K11/Bulldozer with FMA:         8 DP FLOPs/cycle (16 SP FLOPs/cycle)
// IBM Power 7:                        8 DP FLOPs/cycle (16 SP FLOPs/cycle)
//
#pragma omp parallel for default(shared) private(k,k2) firstprivate(a2,ry,rb,rc) reduction(+:ra)
         for(k=0;k<a2;k+=(NA)) {  // 2013-09-11 // 3 instr.
/* gcc-4.1.2 --param max-unroll-times=16 -funroll-loops not unroll NA=10++ ? */
// todo: read in rb rc from file on runtime, avoid compiler-optimiz.
#if 0
// note: not vectorized: relevant stmt not supported: D.4443_128 = D.4439_122 >> 1
           for (k2=0;k2<NA;k2++) { // -funroll-loops * vectorize? = (x2 or x4)
//           ra[k2]=(ra[k2]+rb[k2])*rc[k2]; // gcc44 -msse2 -mavx (bad k10)
             // %4 to save register space
            // NA=1 -O2                      2.41e9/2.41GHz=1x wegoptimiert!?
            // NA=1 -O2 -funroll-loops best! 19.3e9/2.41GHz=8x
// 64bit-benes 10 OPs * 11log2bits = 110 OPs  (old N-loop: 4*N = 256 OPs)
// 40bit-benes 10 OPs * 11log2bits = 110 OPs  (old N-loop: 4*N = 160 OPs)
//     -O0 NA=1 99 (half) clocks (+ lot of moves inside loop)
//     -O1 NA=1 36 (half) clocks (teilweise parallel and,or,shift, no sse2 OPs)
//     -O2 NA=1 30 (half) clocks (teilweise parallel 3andq,2orq,salq,sarq, no sse2 OPs)
// 32bit-butterfly 10 OPs * log2bits = 50 OPs  (old N-loop: 4*N = 128 OPs)
// 16bit-butterfly 10 OPs * log2bits = 40 OPs  (old N-loop: 4*N =  64 OPs)
//  40bit: gcc481-O2..fast NA=1 +/-unroll    80.3e6/s / 2400MHz = 30   clks
//               -O2..fast NA=2 -unroll      92.6e6/s / 2400MHz = 26   clks
//                2movq sarq 3andq salq 2orq (addq a,a statt salq 1 slower!)
// aber per Hand 128bit waere kein problem! +sse2 = 200%
//  32bit: gcc481-O2..fast NA=1 +/-unroll    96.3e6/s / 2400MHz = 25   clks
//  32bit        -O2..fast NA=4 +unroll +sse2 161e6/s / 2400MHz = 15   clks
//  32bit        -O3..fast NA=8 +unroll -sse2 102e6/s / 2400MHz = 23.5 clks
//  32bit        -O3..fast NA=8 +unroll +sse2 290e6/s / 2400MHz =  8.3 clks
//  32bit        -O3..fast NA=12 +unroll +sse2 308e6/s / 2400MHz =  7.8 clks
//           movdqa psrad(>>) 3pand(&) pslld(<<) 2por(|) xmm 5*7OPs
//   weigert sich, bei 2*64bit sse2 zu benutzen
// ToDo: debug using -ftree-vectorizer-verbose=2 
#define R0 ra[k2]
// benes network = butterfly + inverse butterfly (-1stage) == any permutation
// inverse butterfly 6 stages (ketten brauchen nur butterfly?, generator!)
R0=((R0&rb[0])<<1) |((R0>>1) &rb[0])|(R0&~(rb[0]|(rb[0]<<1)));
R0=((R0&rb[1])<<2) |((R0>>2) &rb[1])|(R0&~(rb[1]|(rb[1]<<2)));
R0=((R0&rb[2])<<4) |((R0>>4) &rb[2])|(R0&~(rb[2]|(rb[2]<<4)));
R0=((R0&rb[3])<<8) |((R0>>8) &rb[3])|(R0&~(rb[3]|(rb[3]<<8)));
R0=((R0&rb[4])<<16)|((R0>>16)&rb[4])|(R0&~(rb[4]|(rb[4]<<16)));
R0=((R0&rb[5])<<32)|((R0>>32)&rb[5])|(R0&~(rb[5]|(rb[4]<<32)));
// butterfly 6-1 stages (needed?)
R0=((R0&rb[6])<<16)|((R0>>16)&rb[6])|(R0&~(rb[6]|(rb[6]<<16)));
R0=((R0&rb[7])<<8) |((R0>>8) &rb[7])|(R0&~(rb[7]|(rb[7]<<8)));
R0=((R0&rb[8])<<4) |((R0>>4) &rb[8])|(R0&~(rb[8]|(rb[8]<<4)));
R0=((R0&rb[9])<<2) |((R0>>2) &rb[9])|(R0&~(rb[9]|(rb[9]<<2)));
R0=((R0&rb[10])<<1)|((R0>>1) &rb[10])|(R0&~(rb[10]|(rb[10]<<1)));
//printf(" %08lx\n",ra[0]);
}
#else
//  40bit: gcc481-O2..fast NA=1 +/-unroll    23.8e6/s / 2400MHz = 101 clks
//  32bit: gcc481-O2..fast NA=1 +/-unroll    29.9e6/s / 2400MHz =  80 clks
           for (k2=0;k2<NA;k2++) { // -funroll-loops * vectorize? = (x2 or x4)
          long int j,k=0;
  for (j=0;j<40;j++)
    // 40 * 4 OPs  >> not vectorized! / not vectorized 
    k |= ((ra[k2]>>j)&1)<<(rb[j]);
  // gcc41: not vectorized: unsupported use in stmt.
  // gcc44: vect_model_induction_cost: inside_cost = 1, outside_cost = 2 .
  //   not vectorized: relevant stmt not supported:\
  //            D.4497D.4497_124 = ra$0_512 >> j_505;
//  40bit: gcc44-O2 NA=1 +/-unroll    9.91e6/s / 1200MHz = 121 clks
//  40bit: gcc44-O3 NA=1 +/-unroll    10.9e6/s / 1200MHz = 110 clks
//  40bit: gcc44-O3 NA=1 +/-unroll    11.3e6/s / 1200MHz = 106 clks j+=2
//  40bit: gcc44-O3 NA=1 +/-unroll    12.4e6/s / 1200MHz =  97 clks j+=2 + long
//  40bit: gcc44-O3 NA=1 +/-unroll    15.8e6/s / 1200MHz =  76 clks j+=4 0vec
//  40bit: gcc44-O3 NA=2 +unroll      16.9e6/s / 1200MHz =  71 clks j+=4 0vec
   ra[k2]=k;
}
#endif
//R0=((R0&rb[5])<<32)|((R0>>32)&rb[5])|(R0&~(rb[5]|(rb[5]<<32)));
//           { ra[k2]*=ra[k2]; ra[(k2+NA/2)%NA]+=ry; } // 3.4%,hand=83% NA=12 MIPS
//           { ra[B_k2]*=ra[B_k2]; ra[1^B_k2]+=ry; }   // gcc412 NA=0..9 only! sc
// 0F 95        SETNE r/m8              Set byte if not equal (ZF=0)
// 0F 6F /r 	MOVQ mm, mm/m64 	Move quadword from mm/m64 to mm
         } // for(k=0..a2, k+=2)
         // ToDo: problem x1 x2 ^= over multiple loops hides bitflips!
      } while(--j);               // gcc-4.1 x86 2 instr.
      t2=MPI_Wtime()-t1;
      loops+=i;
      rb[8]=0; for (k2=0;k2<NA;k2++) rb[8]+=ra[k2];
      if (((int)t2)>123456) break;  // easier to find 1234567 at .s code
      MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
      if ((t2>=tmin || t2>60) && !mpi_i) // alive and final message
      /* aBW= 1.02e+06 %8.2e 1%, 2013-06-20 fix bad factor 2 Mem/Task */
      printf("%c%3d %2d %9ld %6.2f %8.2e %8.2e %6.2f %d*+%02d\n",
         ((t2>=tmin)?' ':'#'), mpi_n, omp_n,
      /* factor 4*+ ToDo: aSPEED=threads*loops/maxtime SPEED=loops/mintime */
         loops, t2, mpi_n*a2*1.*loops/t2, a2*1.*loops/(t2*omp_n),
                omp_n*t2*1e9/(loops*1.*a2), NA, tmode);
      for (k2=0;k2<NA  ;k2++) printf(" %08lx", (long)ra[k2]); printf("\n");
      for (k2=0;k2<NA+8;k2++) if (rb[k2]==666) puts("??");
      if (t2>=tmin) break;
    } // i*=2  double loops each round
  } // cnt++ endless
} // test=32

  if (!mpi_i) printf("# memspeed finished\n");
  
  MPI_Finalize();
  return xy;
}
